#include "SDL_events.h"
#include "c_cpp_aliases/aliases.h"
#include "vector/vec.h"
#include "window/window.h"
#include <math.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>

#define ARR_LEN(ARR) sizeof(ARR) / sizeof(ARR[0])

typedef struct {
  f32 radius;
  vec3f_t centre;
  colour_t colour;
} sphere_t;

typedef enum {
  LIGHT_TYPE_POINT,
  LIGHT_TYPE_DIRECTIONAL,
  LIGHT_TYPE_AMBIENT,

  COUNT_LIGHT_TYPE,
} light_type_t;

typedef struct {
  light_type_t type;
  f32 intensity;
  union {
    vec3f_t position;
    vec3f_t direction;
  };
} light_t;

typedef struct {
  sphere_t *spheres;
  light_t *lights;
  u32 spheres_count;
  u32 lights_count;
} scene_t;

typedef struct {
  f32 t1;
  f32 t2;
} solutions_t;

solutions_t ray_intersects_sphere(vec3f_t origin, vec3f_t direction,
                                  sphere_t sphere);
colour_t trace_ray(vec3f_t origin, vec3f_t direction, f32 t_min, f32 t_max,
                   const scene_t *scene, colour_t default_colour);
f32 compute_lighting(vec3f_t P, vec3f_t N, const scene_t *scene);
f32 light_diffuse(f32 light_intensity, vec3f_t light_direction,
                  vec3f_t surface_normal);

i32 main(i32 argc, char *argv[]) {
  colour_t bg =
      (colour_t){.rgba.r = 27, .rgba.g = 38, .rgba.b = 79, .rgba.a = 255};
  vec3f_t camera = {.x = 0.0f, .y = 0.0f, .z = 0.0f};
  vec3f_t viewport = {.x = 1.0f, .y = 1.0f, .z = 1.0f};

  window_t window = {0};

  if (!init_window(&window, 800, 800, "CG From Scratch")) {
    return EXIT_FAILURE;
  }

  bool running = true;
  SDL_Event event = {0};

  sphere_t spheres[] = {
      (sphere_t){
          .radius = 1.0f,
          .centre = (vec3f_t){.x = 0.0f, .y = -1.0f, .z = 3.0f},
          .colour =
              (colour_t){
                  .rgba.r = 245, .rgba.g = 238, .rgba.b = 158, .rgba.a = 255},
      },
      (sphere_t){
          .radius = 1.0f,
          .centre = (vec3f_t){.x = -2.0f, .y = 0.0f, .z = 4.0f},
          .colour =
              (colour_t){
                  .rgba.r = 59, .rgba.g = 142, .rgba.b = 165, .rgba.a = 255},
      },
      (sphere_t){
          .radius = 1.0f,
          .centre = (vec3f_t){.x = 2.0f, .y = 0.0f, .z = 4.0f},
          .colour =
              (colour_t){
                  .rgba.r = 171, .rgba.g = 52, .rgba.b = 40, .rgba.a = 255},
      },
      (sphere_t){
          .radius = 5000.0f,
          .centre = (vec3f_t){.x = 0.0f, .y = -5001.0f, .z = 0.0f},
          .colour =
              (colour_t){
                  .rgba.r = 255, .rgba.g = 255, .rgba.b = 0, .rgba.a = 255},
      },
  };

  light_t lights[] = {
      (light_t){
          .type = LIGHT_TYPE_AMBIENT,
          .intensity = 0.2f,
      },
      (light_t){
          .type = LIGHT_TYPE_POINT,
          .intensity = 0.6f,
          .position = (vec3f_t){.x = 2.0f, .y = 1.0f, .z = 0.0f},
      },
      (light_t){
          .type = LIGHT_TYPE_DIRECTIONAL,
          .intensity = 0.2f,
          .direction = (vec3f_t){.x = 1.0f, .y = 4.0f, .z = 4.0f},
      },
  };

  scene_t scene = {
      .spheres = spheres,
      .lights = lights,
      .spheres_count = ARR_LEN(spheres),
      .lights_count = ARR_LEN(lights),
  };

  i32 w_min = ((i32)window.half_width) * -1;
  i32 w_max = (i32)window.half_width;
  i32 h_min = ((i32)window.half_height) * -1;
  i32 h_max = (i32)window.half_height;

  while (running) {
    while (SDL_PollEvent(&event)) {
      switch (event.type) {
      case SDL_QUIT:
        running = false;
        break;
      }
    }

    clear_window(&window, bg);

    for (i32 y = h_min; y < h_max; ++y) {
      for (i32 x = w_min; x < w_max; ++x) {
        vec3f_t direction = window_to_viewport(&window, x, y, viewport);
        colour_t colour = trace_ray(camera, direction, 1, INFINITY, &scene, bg);
        set_pixel(&window, x, y, colour);
      }
    }

    swap_buffers(&window);
  }

  close_window(&window);

  return EXIT_SUCCESS;
}

solutions_t ray_intersects_sphere(vec3f_t origin, vec3f_t direction,
                                  sphere_t sphere) {
  f32 r = sphere.radius;
  vec3f_t CO = vec_sub(vec3f_t, origin, sphere.centre);

  f32 a = vec_dot(vec3f_t, direction, direction);
  f32 b = 2.0f * vec_dot(vec3f_t, CO, direction);
  f32 c = vec_dot(vec3f_t, CO, CO) - r * r;

  f32 discriminant = b * b - 4 * a * c;
  if (discriminant < 0) {
    return (solutions_t){INFINITY, INFINITY};
  }

  f32 t1 = (-b + sqrtf(discriminant)) / (2 * a);
  f32 t2 = (-b - sqrtf(discriminant)) / (2 * a);

  return (solutions_t){t1, t2};
}

colour_t trace_ray(vec3f_t origin, vec3f_t direction, f32 t_min, f32 t_max,
                   const scene_t *scene, colour_t default_colour) {
  f32 closest_t = INFINITY;
  sphere_t *closest_sphere = NULL;

  for (u32 i = 0; i < scene->spheres_count; ++i) {
    solutions_t solutions =
        ray_intersects_sphere(origin, direction, scene->spheres[i]);

    if (solutions.t1 >= t_min && solutions.t1 <= t_max &&
        solutions.t1 < closest_t) {
      closest_t = solutions.t1;
      closest_sphere = &(scene->spheres[i]);
    }

    if (solutions.t2 >= t_min && solutions.t2 <= t_max &&
        solutions.t2 < closest_t) {
      closest_t = solutions.t2;
      closest_sphere = &(scene->spheres[i]);
    }
  }

  if (!closest_sphere) {
    return default_colour;
  }

  vec3f_t P =
      vec_add(vec3f_t, origin, vec_mul_num(vec3f_t, direction, closest_t));
  vec3f_t N = vec_sub(vec3f_t, P, closest_sphere->centre);
  N = vec_div_num(vec3f_t, N, vec_magnitude(vec3f_t, N));

  f32 light = compute_lighting(P, N, scene);

  return (colour_t){.rgba.r = closest_sphere->colour.rgba.r * light,
                    .rgba.g = closest_sphere->colour.rgba.g * light,
                    .rgba.b = closest_sphere->colour.rgba.b * light,
                    .rgba.a = closest_sphere->colour.rgba.a};
}

f32 compute_lighting(vec3f_t P, vec3f_t N, const scene_t *scene) {
  f32 I = 0.0f;
  light_t light = {0};

  for (u32 i = 0; i < scene->lights_count; ++i) {
    light = scene->lights[i];

    if (light.type == LIGHT_TYPE_AMBIENT) {
      I += light.intensity;
    } else {
      vec3f_t L = {0};

      switch (light.type) {
      case LIGHT_TYPE_POINT:
        L = vec_sub(vec3f_t, light.position, P);
        break;
      case LIGHT_TYPE_DIRECTIONAL:
        L = light.direction;
        break;
      default:
        break;
      }

      I += light_diffuse(light.intensity, L, N);
    }
  }

  return I;
}

f32 light_diffuse(f32 light_intensity, vec3f_t light_direction,
                  vec3f_t surface_normal) {
  f32 dot_product = vec_dot(vec3f_t, light_direction, surface_normal);

  if (dot_product < 0.0f) {
    return 0.0f;
  }

  f32 divisor = vec_magnitude(vec3f_t, light_direction) *
                vec_magnitude(vec3f_t, surface_normal);

  if (divisor == 0.0f) {
    return 0.0f;
  }

  return light_intensity * dot_product / divisor;
}
